import json
import time
import math

import crypko

from requests.adapters import HTTPAdapter

ADDR = '0xca39e90cec69838e73cc4f24ec5077dac44b47d6'
# ITER_WEIGHT = 1/10
# SIM_THRESHOLD = 12

CAP = 270000
USE_CAP = False

ATTRIBS = 'cooldownready'
FILTERS = None


# FILTERS = 'iteration:~3'  # Iter 3 or under only
# FILTERS = 'cooldown:ur'  # Only ultra rapids


def score(c1, c2):
    if USE_CAP and c1.id > CAP and c2.id > CAP:
        return 0

    if c1.auction.active or c2.auction.active:
        return 0

    similarity = bin(c1.attrs.attribute_int & c2.attrs.attribute_int).count('1')
    n_sim = bin(int(c1.noise) & int(c2.noise)).count('1')

    # if similarity < SIM_THRESHOLD:
    #     return 0
    # return similarity / (max(c1.iteration, c2.iteration) ** ITER_WEIGHT)

    return similarity * n_sim


def legal(c1, c2, cache):
    a = [c1.id, c2.id, *cache[c1.id], *cache[c2.id]]
    return len(a) == len(set(a))


def load_cache():
    try:
        with open('cache.json') as file_:
            dat = json.load(file_)
    except FileNotFoundError:
        dat = {}

    rtn = {}
    for i in dat:
        rtn[int(i)] = (*map(int, dat[i]),)

    return rtn


def save_cache(dat):
    with open('cache.json', 'w') as file_:
        json.dump(dat, file_)


def best_pair(crypkos, cooldown=False):
    scores = []
    cache = load_cache()

    for c1 in crypkos:
        if c1.id not in cache:
            c1.ensure_complete()
            cache[c1.id] = (c1.matron.id if c1.matron is not None else time.time(),
                            c1.sire.id if c1.sire is not None else time.time())
            save_cache(cache)

        for c2 in crypkos:
            if c2.id not in cache:
                c2.ensure_complete()
                cache[c2.id] = (c2.matron.id if c2.matron is not None else time.time(),
                                c2.sire.id if c2.sire is not None else time.time())
                save_cache(cache)

            if not legal(c1, c2, cache):
                continue

            scores.append((c1, c2, score(c1, c2)))

    scores.sort(key=lambda x: x[2], reverse=True)
    if scores:
        return scores[0][0], scores[0][1]
    return None


def purge_attributes(api, attribs):
    print(f'==> Collecting crypkos for \'{attribs}\'')
    crypkos = [i for i in api.search(owner_addr=ADDR, attributes=attribs)[1] if not i.auction.active]
    txs = []
    for crypko in crypkos:
        print(f'==> Selling #{crypko.id}')
        tx = api.contract.start_auc(crypko.id)
        print(f' ::  TX {tx.hex()}')
        txs.append(tx)

    print(f'==> Flushing {len(txs)} transactions')
    for tx in txs:
        print(f' ::  TX {tx.hex()}')
        api.contract.wait_for_tx(tx)
    print(f'==> Done!')


def main():
    with open('priv.key') as key_file:
        api = crypko.API(ADDR, key_file.read().strip())

    batch_size = 24
    while True:
        purge_attributes(api, 'dark skin')
        purge_attributes(api, 'glasses')

        print(f'==> Requesting crypkos..')
        my_crypkos = [i for i in api.search(owner_addr=ADDR,
                                            attributes=ATTRIBS,
                                            filters=FILTERS)[1] if not i.auction.active]

        print(f' ::  {len(my_crypkos)} usable')

        txs = []
        for _ in range(batch_size):
            print(f'==> Locating best pair..')
            pair = best_pair(my_crypkos)
            if pair is None:
                print(' ::  No legal fuses avaliable. Breaking.')
                break

            print(f' ::  Fusing #{pair[0].id} with #{pair[1].id}..')
            tx = api.contract.fuse(pair[0].id, pair[1].id)
            print(f' ::  Transaction {tx.hex()}')
            txs.append(tx)

            my_crypkos.remove(pair[0])
            my_crypkos.remove(pair[1])
        print(f'==> Flushing {len(txs)} transactions')
        for tx in txs:
            print(f' ::  TX {tx.hex()}')
            api.contract.wait_for_tx(tx)
        print(f' ::  Done!')


if __name__ == '__main__':
    main()



# Copyright (c) 2018 Bottersnike
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import requests
import hashlib

from .errors import CrypkoApiError

IMAGE_CODE = 'asdasd3edwasd'

ATTR_MAP = {'hair': ['blonde hair', 'brown hair', 'black hair', 'blue hair', 'pink hair', 'purple hair', 'green hair',
                     'red hair', 'silver hair', 'white hair', 'orange hair', 'aqua hair', 'grey hair'],
            'eyes': ['blue eyes', 'red eyes', 'brown eyes', 'green eyes', 'purple eyes', 'yellow eyes', 'pink eyes',
                     'aqua eyes', 'black eyes', 'orange eyes'],
            'hair_style': ['long hair', 'short hair', 'twintails', 'drill hair', 'ponytail'],
            'attribute': ['dark skin', 'blush', 'smile', 'open mouth', 'hat', 'ribbon', 'glasses']}


class CrypkoAttributes:
    def __init__(self, attr_string):
        self.attribute_int = int(attr_string)

    @property
    def attributes(self):
        attrs = bin(self.attribute_int)[2:].rjust(38, '0')

        rtn = [
            (ATTR_MAP['hair'][int(attrs[30: 34], 2) % len(ATTR_MAP['hair'])], True),
            (ATTR_MAP['eyes'][int(attrs[22: 26], 2) % len(ATTR_MAP['eyes'])], True),
            (ATTR_MAP['hair_style'][int(attrs[14: 18], 2) % len(ATTR_MAP['hair_style'])], True)
        ]

        for n, i in enumerate(ATTR_MAP['attribute'][::-1]):
            rtn.append(
                (i, bool(int(attrs[n * 2])))
            )

        return rtn

    @property
    def sub_attributes(self):
        attrs = bin(self.attribute_int)[2:].rjust(38, '0')

        rtn = [
            (ATTR_MAP['hair'][int(attrs[34: 38], 2) % len(ATTR_MAP['hair'])], True),
            (ATTR_MAP['eyes'][int(attrs[26: 30], 2) % len(ATTR_MAP['eyes'])], True),
            (ATTR_MAP['hair_style'][int(attrs[18: 22], 2) % len(ATTR_MAP['hair_style'])], True)
        ]

        for n, i in enumerate(ATTR_MAP['attribute'][::-1]):
            rtn.append(
                (i, bool(int(attrs[n * 2 + 1])))
            )

        return rtn


class CrypkoOwner:
    def __init__(self, data):
        self.avatar_dat = data.get('avatarCrypko')
        self.username = data.get('username')
        self.address = data.get('address')
        self.avatar = data.get('avatar')
        self.bio = data.get('introduction')


class CrypkoAuction:
    SALE = 0
    RENTAL = 1

    def __init__(self, start_price, end_price, duration, started, ends, sale, rental):
        self.start_price = start_price
        self.end_price = end_price
        self.duration = duration
        self.started = started
        self.ends = ends

        self.active = sale or rental

        self.type = self.SALE if sale else self.RENTAL if rental else None


class Crypko:
    def __init__(self, data, api):
        self._api = api

        self.sire = self.matron = self.owner = None

        self.complete = False

        self._load_details(data)

    def _load_details(self, data):
        self.id = data.pop('id')

        # Attribute parsing
        attrs = CrypkoAttributes(data.pop('attrs', 0))
        self.sub_attributes = attrs.sub_attributes
        self.attributes_int = attrs.attribute_int
        self.attributes = attrs.attributes

        # Load auction details
        start_price = int(data.pop('startingPrice', 0))
        end_price = int(data.pop('endingPrice', 0))
        duration = int(data.pop('duration', 0))
        started = int(data.pop('startedAt', 0))
        ends = int(data.pop('endAt', 0))

        on_sale = data.pop('onSaleClockAuction', False)
        on_rental = data.pop('onRentalClockAuction', False)

        self.auction = CrypkoAuction(start_price, end_price, duration, started, ends, on_sale, on_rental)

        # Metadata
        self.derivatives = data.pop('derivatives', [])
        self.likes = data.pop('likeCount', 0)
        self.raw_noise = int(data.pop('noise', 0))

        self.iteration = data.pop('iteration', 0)

        self.cooldown_left = data.pop('cooldownRemain', 0)
        self.cooldown_index = data.pop('cooldownIndex')

        if 'owner' in data:
            self.owner = CrypkoOwner(data.pop('owner'))
            data.pop('ownerAddr', None)
        elif 'ownerAddr' in data:
            self.owner = CrypkoOwner({'address': data.pop('ownerAddr')})

        self.bio = data.pop('bio', None)
        self.name = data.pop('name', None)

        self.next_action_at = data.pop('nextActionAt', 0)

        # Visual attributes
        # # Features
        self.hair_colour = data.pop('hairColor', '')[:-5] or None
        self.eye_colour = data.pop('eyeColor', '')[:-5] or None
        self.hair_style = data.pop('hairStyle', None)
        self.open_mouth = data.pop('openMouth', None)
        self.dark_skin = data.pop('darkSkin', None)
        self.blush = data.pop('blush', None)
        # # Objects
        self.glasses = data.pop('glasses', None)
        self.ribbon = data.pop('ribbon', None)
        self.smile = data.pop('smile', None)
        self.hat = data.pop('hat', None)

        self.updated_block_num = data.pop('updatedBlockNum', '')

        self.last_liked = data.pop('lastLiked', '')  # TODO: datetime

        # Genetics
        if 'sire' in data:
            self.sire = Crypko(data.pop('sire'), self._api)
            data.pop('sireId', None)
        elif 'sireId' in data:
            self.sire = Crypko({'id': data.pop('sireId')}, self._api)

        if 'matron' in data:
            self.matron = Crypko(data.pop('matron'), self._api)
            data.pop('matronId', None)
        elif 'matronId' in data:
            self.matron = Crypko({'id': data.pop('matronId')}, self._api)

        if data:
            print('[WARNING] Unhandled Crypko data: {}'.format(data))

    def __str__(self):
        rtn = 'Crypko #{}'.format(self.id)
        if self.name is not None:
            rtn += ' (' + self.name + ')'
        if self.auction.active:
            if self.auction.type == CrypkoAuction.SALE:
                rtn += ' (on sale)'
            else:
                rtn += ' (for rent)'

        return rtn

    def __eq__(self, other):
        return other.id == self.id

    @property
    def noise(self):
        noise = bin(self.raw_noise)[2:].rjust(256, '0')
        return [int(noise[n - 2: n], 2) for n in range(len(noise), 0, -2)]

    def reload_details(self):
        res = requests.get('{}crypkos/{}/detail'.format(self._api.DOMAIN, self.id))

        if res.status_code != 200:
            raise CrypkoApiError('Got response code {}. ({})'.format(res.status_code, res.text))

        self._load_details(res.json())
        self.complete = True

    def ensure_complete(self):
        if not self.complete:
            self.reload_details()

    @property
    def image(self):
        return 'https://img.crypko.ai/daisy/' + \
               hashlib.sha1((str(self.noise) + IMAGE_CODE + str(self.attrs.attribute_int)).encode()).hexdigest() + \
               '_lg.jpg'

    @property
    def image_sm(self):
        return self.image[:-6] + 'sm.jpg'



# Copyright (c) 2018 Bottersnike
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import threading
import math

import requests

from .errors import CrypkoApiError, CrypkoNotFoundError
from .blockchain import ContractHandler
from .objects import Crypko


class API:
    DOMAIN = 'https://api.crypko.ai/'
    PER_PAGE = 12  # Can't change this :(

    def __init__(self, address=None, key=None):
        if key is not None and address is not None:
            self._contract = ContractHandler(key, address)
        else:
            self._contract = None

    def _search(self, owner_addr=None, page=None, sort=None, category=None, attributes=None, filters=None):
        req = requests.get(self.DOMAIN + 'crypkos/search', params={
            'ownerAddr': owner_addr,
            'page': page,
            'sort': sort,
            'category': category,
            'attributes': attributes,
            'filters': filters
        })
        if req.status_code != 200:
            raise CrypkoApiError('Got response code {}. ({})'.format(req.status_code, req.text))

        return req.json()

    @property
    def contract(self):
        if self._contract is None:
            raise CrypkoApiError('Contract not loaded. Did you provide an address and key?')
        return self._contract

    def get_crypko(self, crypko_id):
        res = requests.get('{}crypkos/{}/detail'.format(self.DOMAIN, crypko_id))

        if res.status_code == 404:
            raise CrypkoNotFoundError
        if res.status_code != 200:
            raise CrypkoApiError('Got response code {}. ({})'.format(res.status_code, res.text))

        return Crypko(res.json(), self)

    def threaded_search(self, callback, threads=8, start=0, results=-1, **kwargs):
        if results == 0:
            return
        start = max(0, start)
        page = math.floor(start / self.PER_PAGE) + 1

        kwargs.pop('page', None)
        if results < 0:
            r = self._search(**kwargs)
            results = r['totalMatched'] - start + 1

        page_end = math.ceil((start + results) / self.PER_PAGE) + 1

        def runner():
            nonlocal page
            while page < page_end:
                me_page = page
                page += 1
                res = self._search(page=me_page, **kwargs)

                if not res['crypkos']:
                    break
                
                for n, i in enumerate(res['crypkos']):
                    if n + (page - 1) * self.PER_PAGE >= start:
                        if n + (page - 1) * self.PER_PAGE < start + results:
                            callback(Crypko(i, self))

        pool = []
        for _ in range(threads):
            pool.append(threading.Thread(target=runner))
            pool[-1].start()

        for thread in pool:
            thread.join()

    def search(self, start=0, results=-1, **kwargs):
        """
        Search for Crypkos. Will retrieve `results` Crypkos starting at `start`.
        If results is negative, we will read until the end of results.

        TODO: Document kwargs
        """
        if results == 0:
            return
        start = max(0, start)
        page = math.floor(start / self.PER_PAGE) + 1
        kwargs.pop('page', None)

        result = self._search(page=page, **kwargs)

        def iters():
            nonlocal result, page
            results_sent = 0
            while results < 0 or results_sent < results:
                if not result['crypkos']:
                    break

                if results_sent == 0:
                    result['crypkos'] = result['crypkos'][start % self.PER_PAGE:]

                for i in result['crypkos']:
                    if results < 0 or results_sent < results:
                        yield Crypko(i, self)
                        results_sent += 1
                page += 1

                if results < 0 or results_sent < results:
                    result = self._search(page=page, **kwargs)
        
        return result['totalMatched'], iters()



import crypko


ADDR = '0xca39e90cec69838e73cc4f24ec5077dac44b47d6'
SELL_DOWN_TO = 251824

GIFT = True

CAN_FIRST = 100
GIVE_TO = '0xb24978d166e766d9557fa75d9b7b16f0beccebc2'

START_PRICE = 0.8
END_PRICE = 0.6


def main():
    with open('priv.key') as key_file:
        api = crypko.API(ADDR, key_file.read().strip())

    txs = []
    count, crypkos = api.search(owner_addr=ADDR)

    cnt = 0
    for c in crypkos:
        if GIFT and not c.auction.active:
            if c.id not in [366013] and c.id > SELL_DOWN_TO:
                print(f'==> Gifting #{c.id}')

                tx = api.contract.gift(c.id, GIVE_TO)
                print(f' ::  TX {tx.hex()}')
                txs.append(tx)

                cnt += 1

                if cnt >= CAN_FIRST:
                    break
        elif not GIFT and c.id > SELL_DOWN_TO and not c.on_sale:
            print(f'==> Selling #{c.id}')

            tx = api.contract.start_auc(c.id, START_PRICE, END_PRICE)
            print(f' ::  TX {tx.hex()}')
            txs.append(tx)

    print(f'==> Flushing {len(txs)} transactions')
    for tx in txs:
        print(f' ::  TX {tx.hex()}')
        api.contract.wait_for_tx(tx)

    print(f'==> Done!')


if __name__ == '__main__':
    main()



# Copyright (c) 2018 Bottersnike
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This example demonstrates how to use the threaded search to rapidly
# download the image for every crypko a user owns. It relies of a directory
# named out/ existing already.

import requests
import crypko

from requests.adapters import HTTPAdapter


THREADS = 16  # The number of seperate threads to run
# Replace this with the address of the user to download from
ADDRESS = '0xCa39E90CeC69838e73CC4F24Ec5077daC44B47d6'


def prepare_callback(session):
    def callback(cryp):
        print('Processing crypkoi#%d' % cryp.id)

        # cryp.image could be changes to cryp.image_sm to download the smaller
        # 192x192 version of the image instead of the 512x512 version.
        r = session.get(cryp.image, stream=True)

        # Download and save the image
        with open('out/%d.jpg' % cryp.id, 'wb') as file_:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    file_.write(chunk)

    return callback


def main():
    api = crypko.API()

    # Create a custom session so we can set a retry count
    session = requests.Session()
    session.mount('https://', HTTPAdapter(max_retries=5))

    # Perform the search
    api.threaded_search(prepare_callback(session), threads=THREADS, owner_addr=ADDRESS)


if __name__ == '__main__':
    main()




import os

from setuptools import setup


version = __import__('crypko').__version__


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


setup(
    name='crypko.py',
    version=version,
    author='Bottersnike',
    author_email='bottersnik237@gmail.com',
    description=('A basic wrapper around the Crypko platform.'),
    long_description=read('README.rst'),
    license='MIT',
    install_requires=[
        'requests',
        'web3'
    ],
    packages=['crypko', 'crypko.examples'],
    url='https://gitlab.com/Bottersnike/crypko.py',
    project_urls={
        "Bug Tracker": "https://gitlab.com/Bottersnike/crypko.py/issues",
        "Source Code": "https://gitlab.com/Bottersnike/crypko.py",
    },
    package_data={
        'crypko': ['abi.json'],
    },
)




# Copyright (c) 2018 Bottersnike
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os

from web3.middleware import geth_poa_middleware
from web3 import Web3, HTTPProvider


class ContractHandler:
    NODE = 'https://mainnet.infura.io/metamask'
    NODE = 'https://rinkeby.infura.io/metamask'  # Remove when switched to mainnet
    CONTRACT_ADDRESS = '0xFe2149773B3513703E79Ad23D05A778A185016ee'

    SPECIAL_ADDRS = {
        'ClockAuction': '0xC8cD5cA4f1a91cdb103EC19d3B46a1E42A862025'
    }

    RECEPIENT = '0xfe2149773b3513703e79ad23d05a778a185016ee'

    def __init__(self, key, address):
        self.web3 = Web3(HTTPProvider(self.NODE))
        self.web3.middleware_stack.inject(geth_poa_middleware, layer=0)

        # Load the ABI
        abi_path = os.path.join(os.path.dirname(__file__), 'abi.json')
        with open(abi_path) as file_:
            abi = json.load(file_)

        self.contracts = {}
        for i in abi:
            self.contracts[i] = self.web3.eth.contract(
                    address=self.SPECIAL_ADDRS.get(i, self.CONTRACT_ADDRESS),
                    abi=abi[i])

        # Variables for later
        self.priv_key = key
        self.address = address
        self.nonce = None

    def _send(self, abi, value=0):
        if self.nonce is None:
            self.nonce = self.web3.eth.getTransactionCount(Web3.toChecksumAddress(self.address))
        else:
            self.nonce += 1

        tx = { 
            'nonce': self.nonce,
            'gasPrice': Web3.toWei(1, 'gwei'),
            'from': Web3.toChecksumAddress(self.address),
            'gas': 1000000000,
            'value': value,
        }   
        gas = abi.estimateGas(tx)
        tx['gas'] = int(gas * 1.3)
        tx = abi.buildTransaction(tx)
        tx = self.web3.eth.account.signTransaction(tx, self.priv_key)

        return self.web3.eth.sendRawTransaction(tx.rawTransaction)

    def start_auc(self, crypko, start=0.7, end=0.4, duration=172800):
        start = Web3.toWei(start, 'ether')
        end = Web3.toWei(end, 'ether')
        start, end = int(start), int(end)

        create = self.contracts['CrypkoCore'].functions.createSaleAuction(crypko, start, end, duration)

        return self._send(create)

    def bid(self, crypko):
        cost = self.contracts['ClockAuction'].getCurrentPrice(crypko).call()
        bid = self.contracts['ClockAuction'].functions.bid(crypko)

        return self._send(bid, cost)

    def fuse(self, first, second):
        fee = self.contracts['CrypkoFusion'].call().fusionFee()
        fwa = self.contracts['CrypkoFusion'].functions.fuseWithAuto(first, second)

        return self._send(fwa, fee)

    def cancel(self, crypko):
        cancel = self.contracts['ClockAuction'].functions.cancelAuction(crypko)

        return self._send(cancel)

    def gift(self, crypko, to):
        to = Web3.toChecksumAddress(to)
        gift = self.contracts['CrypkoAuction'].functions.transfer(to, crypko)

        return self._send(gift)

    def wait_for_tx(self, tx):
        return self.web3.eth.waitForTransactionReceipt(tx)
